import logging

logger = logging.getLogger(__name__)


class Pipeline(object):
    def __init__(self, name, *phases):
        """

        Parameters
        ----------
        name: str
            The name of this pipeline
        phases: [ph.Phase]
            Phases
        """
        self.name = name
        self.phases = phases

    def run(self, image):
        from autolens.pipeline import phase as ph
        results = []
        for i, phase in enumerate(self.phases):
            logger.info("Running Phase {} (Number {})".format(phase.name, i))
            results.append(phase.run(image, ph.ResultsCollection(results)))
        return results

    def __add__(self, other):
        """
        Compose two pipelines

        Parameters
        ----------
        other: Pipeline
            Another pipeline

        Returns
        -------
        composed_pipeline: Pipeline
            A pipeline that runs all the  phases from this pipeline and then all the phases from the other pipeline
        """
        return Pipeline("{} + {}".format(self.name, other.name), *(self.phases + other.phases))



"""
autolens

Usage:
  autolens download_config
  autolens pipeline
  autolens pipeline <name> (--image=<image>) (--pixel-scale=<pixel-scale>) [--config=<config>] [--output=<output>]
  autolens -h | --help
  autolens --version

Options:
  -h --help                         Show this screen.
  --version                         Show version.
  --pixel-scale=<pixel-scale>       The scale of one pixel in the image.
  --image=<image>                   The path to the folder that contains the image data.
  --config=<config>                 The path to the folder that contains the configuration [default: config].
  --output=<output>                 The path to the folder where data should be output [default: output].

Examples:
  autolens download_config
  autolens pipeline
  autolens pipeline profile --image=hst_0/ --pixel-scale=0.05
  autolens pipeline profile --image=hst_0 --pixel-scale=0.05 --output=output_folder --config=config_folder

Help:
  For help using this tool, please open an issue on the Github repository:
  https://github.com/Jammy2211/PyAutoLens
"""

from inspect import getmembers, isclass

from docopt import docopt

from . import __version__


def main():
    """Main CLI entrypoint."""
    import autolens.commands

    options = docopt(__doc__, version=__version__)

    # Here we'll try to dynamically match the command the user is trying to run
    # with a pre-defined command class we've already created.
    for (k, v) in options.items():
        if hasattr(autolens.commands, k) and v:
            module = getattr(autolens.commands, k)
            command = [command[1] for command in getmembers(module, isclass) if command[0] != 'Base'][0]
            command = command(options)
            command.run()


if __name__ == "__main__":
    main()



"""
Analyse only the source galaxy.

This pipeline fits the source galaxy with a rectangular pixelization. It assumes the foreground galaxy's mass profile
comprises Spherical Isothermal and External Shear components.
"""

name = "source"


def make():
    from autolens.pipeline import phase as ph
    from autolens.pipeline import pipeline as pl
    from autolens.analysis import galaxy_prior as gp
    from autolens.pixelization import pixelization as px
    from autolens.imaging import mask as msk
    from autolens.profiles import light_profiles, mass_profiles

    # 1) Mass: SIE+Shear
    #    Source: Sersic
    #    NLO: LM
    def mask_function(img):
        return msk.Mask.circular(img.shape_arc_seconds, img.pixel_scale, 2)

    phase1 = ph.SourceLensPhase(
        lens_galaxy=gp.GalaxyPrior(
            sie=mass_profiles.SphericalIsothermal,
            shear=mass_profiles.ExternalShear),
        source_galaxy=gp.GalaxyPrior(
            sersic=light_profiles.EllipticalSersic),
        mask_function=mask_function)

    # 2) Mass: SIE+Shear (priors from phase 1)
    #    Source: 'smooth' pixelization (include regularization parameter(s) in the model)
    #    NLO: LM
    class PriorLensPhase(ph.PixelizedSourceLensPhase):
        def pass_priors(self, previous_results):
            self.lens_galaxy = previous_results.last.variable.lens_galaxy

    phase2 = PriorLensPhase(pixelization=px.RectangularRegConst,
                            mask_function=mask_function)

    class ConstantLensPhase(ph.PixelizedSourceLensPhase):
        def pass_priors(self, previous_results):
            self.lens_galaxy = previous_results.last.constant.lens_galaxy

    phase3 = ConstantLensPhase(pixelization=px.RectangularRegConst,
                               mask_function=mask_function)
    return pl.Pipeline(name, phase1, phase2, phase3)



from autolens.profiles import light_profiles, mass_profiles


class EllipticalSersicMassAndLight(light_profiles.EllipticalSersic, mass_profiles.EllipticalSersicMass):
    def __init__(self):
        pass



from autolens.pipeline import profile_pipeline
from autolens.pipeline import source_only_pipeline

pipeline_dict = {}


def add(module):
    pipeline_dict[module.name] = module.make


add(profile_pipeline)
add(source_only_pipeline)



from autolens.commands.base import Base, current_directory
from autolens import conf


class Pipeline(Base):

    def run(self):
        from autolens import pipeline
        name = self.options['<name>']
        if name is not None:
            if name not in pipeline.pipeline_dict:
                print("No pipeline called '{}' found".format(name))
                return
            conf.instance = conf.Config(self.config_path, self.output_path)
            self.run_pipeline(pipeline.pipeline_dict[name]())

        print("Available Pipelines:\n")
        print("\n".join(list(pipeline.pipeline_dict.keys())))

    def run_pipeline(self, pipeline):
        pipeline.run(self.load_image())

    @property
    def image_path(self):
        image_path = self.options['--image']
        if image_path is None:
            print("Please specify the path to the image folder")
            return
        if not image_path.startswith("/"):
            image_path = "{}/{}".format(current_directory, image_path)
        return image_path

    @property
    def pixel_scale(self):
        return float(self.options['--pixel-scale'])

    @property
    def config_path(self):
        config_path = self.options['--config']
        if not conf.is_config(config_path):
            print("No config found at {}. Try running 'autolens download_config'".format(config_path))
            exit(1)
        return config_path

    @property
    def output_path(self):
        output_path = self.options['--output']
        if not output_path.startswith("/"):
            output_path = "{}/{}".format(current_directory, output_path)
        return output_path

    def load_image(self):
        from autolens.imaging import scaled_array
        from autolens.imaging import image as im

        data = scaled_array.ScaledArray.from_fits(file_path='{}/image'.format(self.image_path), hdu=0,
                                                  pixel_scale=self.pixel_scale)
        noise = scaled_array.ScaledArray.from_fits(file_path='{}/noise'.format(self.image_path), hdu=0,
                                                   pixel_scale=self.pixel_scale)
        psf = im.PSF.from_fits(file_path='{}/psf'.format(self.image_path), hdu=0)

        return im.Image(array=data, pixel_scale=self.pixel_scale, psf=psf, noise=noise)



"""
Analyse a lens source system using only profiles.

This pipeline fits the source light with an Elliptical Sersic profile, the lens light with an Elliptical Sersic profile
and the lens mass with a Spherical Isothermal profile.
"""


name = "profile"


def make():
    from autolens.pipeline import phase as ph
    from autolens.pipeline import pipeline as pl
    from autolens.autopipe import non_linear as nl
    from autolens.analysis import galaxy_prior as gp
    from autolens.imaging import mask as msk
    from autolens.profiles import light_profiles, mass_profiles

    optimizer_class = nl.MultiNest

    # 1) Lens Light: EllipticalSersic
    #    Mass: None
    #    Source: None
    #    NLO: MultiNest
    #    Image: Observed Image
    #    Mask: Circle - 3.0"
    phase1 = ph.LensOnlyPhase(lens_galaxy=gp.GalaxyPrior(elliptical_sersic=light_profiles.EllipticalSersic),
                              optimizer_class=optimizer_class, name="{}/phase1".format(name))

    class LensSubtractedPhase(ph.SourceLensPhase):
        def modify_image(self, masked_image, previous_results):
            return masked_image - previous_results.last.lens_galaxy_image

        def pass_priors(self, previous_results):
            self.lens_galaxy.sie.centre = previous_results.last.variable.lens_galaxy.elliptical_sersic.centre

    # 2) Lens Light : None
    #    Mass: SIE (use lens light profile centre from previous phase as prior on mass profile centre)
    #    Source: EllipticalSersic
    #    NLO: MultiNest
    #    Image: Lens Subtracted Image (previous phase)
    #    Mask: Annulus (0.4" - 3.0")
    def annular_mask_function(img):
        return msk.Mask.annular(img.shape_arc_seconds, pixel_scale=img.pixel_scale, inner_radius=0.4,
                                outer_radius=3.)

    phase2 = LensSubtractedPhase(lens_galaxy=gp.GalaxyPrior(sie=mass_profiles.SphericalIsothermal),
                                 source_galaxy=gp.GalaxyPrior(elliptical_sersic=light_profiles.EllipticalSersic),
                                 optimizer_class=optimizer_class,
                                 mask_function=annular_mask_function,
                                 name="{}/phase2".format(name))

    # 3) Lens Light : Elliptical Sersic (Priors phase 1)
    #    Mass: SIE (Priors phase 2)
    #    Source: Elliptical Sersic (Priors phase 2)
    #    NLO: MultiNest
    #    Image: Observed Image
    #    Mask: Circle - 3.0"
    class CombinedPhase(ph.SourceLensPhase):
        def pass_priors(self, previous_results):
            self.lens_galaxy = gp.GalaxyPrior(
                elliptical_sersic=previous_results.first.variable.lens_galaxy.elliptical_sersic,
                sie=previous_results.last.variable.lens_galaxy.sie)
            self.source_galaxy = previous_results.last.variable.source_galaxy

    phase3 = CombinedPhase(optimizer_class=optimizer_class,
                           name="{}/phase3".format(name))
    # 3H) Hyper-Parameters: Make Lens Galaxy and Source Galaxy Hyper-Galaxies.
    #     Lens Light / Mass / Source - Fix parameters to phase 3 most likely result
    #     NLO: DownhillSimplex
    #     Image: Observed Image
    #     Mask: Circle - 3.0"
    phase3h = ph.SourceLensHyperGalaxyPhase(name="{}/phase3h".format(name))

    # 4) Repeat phase 3, using its priors and the hyper-galaxies fixed to their optimized values.
    #    Lens Light: Elliptical Sersic (Priors phase 3)
    #    Mass: SIE (Priors phase 3)
    #    Source: Elliptical Sersic (Priors phase 3)
    #    NLO: MultiNest
    #    Image: Observed Image
    #    Mask: Circle - 3.0"
    class CombinedPhase2(ph.SourceLensPhase):
        def pass_priors(self, previous_results):
            phase_3_results = previous_results[2]
            self.lens_galaxy = phase_3_results.variable.lens_galaxy
            self.source_galaxy = phase_3_results.variable.source_galaxy
            self.lens_galaxy.hyper_galaxy = previous_results.last.constant.lens_galaxy.hyper_galaxy
            self.source_galaxy.hyper_galaxy = previous_results.last.constant.source_galaxy.hyper_galaxy

    phase4 = CombinedPhase2(optimizer_class=optimizer_class, name="{}/phase4".format(name))
    return pl.Pipeline("profile_pipeline", phase1, phase2, phase3, phase3h, phase4)
